@@ -4,11 +4,9 @@ Click command to do static annotation searching via Stevedore plugins.
 import datetime
 import errno
 import os
-import pprint
 import re
 from abc import ABCMeta, abstractmethod
 
-import six
 import yaml
 from stevedore import named
 
@@ -16,12 +14,12 @@ from code_annotations.exceptions import ConfigurationException
 from code_annotations.helpers import VerboseEcho
 
 
-class AnnotationConfig(object):
+class AnnotationConfig:
     """
     Configuration shared among all Code Annotations commands.
     """
 
-    def __init__(self, config_file_path, report_path_override, verbosity, source_path_override=None):
+    def __init__(self, config_file_path, report_path_override=None, verbosity=1, source_path_override=None):
         """
          Initialize AnnotationConfig.
 
@@ -33,6 +31,7 @@ class AnnotationConfig(object):
         """
         self.groups = {}
         self.choices = {}
+        self.optional_groups = []
         self.annotation_tokens = []
         self.annotation_regexes = []
         self.mgr = None
@@ -41,33 +40,27 @@ class AnnotationConfig(object):
         self.echo = VerboseEcho()
 
         with open(config_file_path) as config_file:
-            raw_config = yaml.load(config_file)
+            raw_config = yaml.safe_load(config_file)
 
         self._check_raw_config_keys(raw_config)
 
         self.safelist_path = raw_config['safelist_path']
         self.extensions = raw_config['extensions']
 
-        try:
-            self.coverage_target = float(raw_config['coverage_target'])
-        except (TypeError, ValueError):
-            raise ConfigurationException(
-                'Coverage target must be a number between 0 and 100 not "{}".'.format(raw_config['coverage_target'])
-            )
-
-        if self.coverage_target < 0.0 or self.coverage_target > 100.0:
-            raise ConfigurationException(
-                'Invalid coverage target. {} is not between 0 and 100.'.format(self.coverage_target)
-            )
-
         self.verbosity = verbosity
         self.echo.set_verbosity(verbosity)
 
         self.report_path = report_path_override if report_path_override else raw_config['report_path']
-        self.echo("Configured for report path: {}".format(self.report_path))
+        self.echo(f"Configured for report path: {self.report_path}")
 
         self.source_path = source_path_override if source_path_override else raw_config['source_path']
-        self.echo("Configured for source path: {}".format(self.source_path))
+        self.echo(f"Configured for source path: {self.source_path}")
+
+        self._configure_coverage(raw_config.get('coverage_target', None))
+        self.report_template_dir = raw_config.get('report_template_dir')
+        self.rendered_report_dir = raw_config.get('rendered_report_dir')
+        self.rendered_report_file_extension = raw_config.get('rendered_report_file_extension')
+        self.rendered_report_source_link_prefix = raw_config.get('rendered_report_source_link_prefix')
 
         self._configure_annotations(raw_config)
         self._configure_extensions()
@@ -83,7 +76,7 @@ class AnnotationConfig(object):
             ConfigurationException on any missing keys
         """
         errors = []
-        for k in ('report_path', 'source_path', 'safelist_path', 'annotations', 'extensions', 'coverage_target'):
+        for k in ('report_path', 'source_path', 'safelist_path', 'annotations', 'extensions'):
             if k not in raw_config:
                 errors.append(k)
 
@@ -116,11 +109,23 @@ class AnnotationConfig(object):
         Returns:
             True if the type of the annotation is correct for a choice group, otherwise False
         """
-        return isinstance(token_or_group, dict)
+        return isinstance(token_or_group, dict) and "choices" in token_or_group
+
+    def _is_optional_group(self, token_or_group):
+        """
+        Determine if an annotation is an optional group.
+
+        Args:
+            token_or_group: The annotation being checked
+
+        Returns:
+            True if the annotation is optional, otherwise False.
+        """
+        return isinstance(token_or_group, dict) and bool(token_or_group.get("optional"))
 
     def _is_annotation_token(self, token_or_group):
         """
-        Determine if an annotation is a free-form text type.
+        Determine if an annotation has the right format.
 
         Args:
             token_or_group: The annotation being checked
@@ -128,7 +133,42 @@ class AnnotationConfig(object):
         Returns:
             True if the type of the annotation is correct for a text type, otherwise False
         """
-        return token_or_group is None
+        if token_or_group is None:
+            return True
+        if isinstance(token_or_group, dict):
+            # If annotation is a dict, only a few keys are tolerated
+            return set(token_or_group.keys()).issubset({"choices", "optional"})
+        return False
+
+    def _add_annotation_token(self, token):
+        if token in self.annotation_tokens:
+            raise ConfigurationException(f'{token} is configured more than once, tokens must be unique.')
+        self.annotation_tokens.append(token)
+
+    def _configure_coverage(self, coverage_target):
+        """
+        Set coverage_target to the specified value.
+
+        Args:
+            coverage_target:
+
+        Returns:
+
+        """
+        if coverage_target:
+            try:
+                self.coverage_target = float(coverage_target)
+            except (TypeError, ValueError) as error:
+                raise ConfigurationException(
+                    f'Coverage target must be a number between 0 and 100 not "{coverage_target}".'
+                ) from error
+
+            if self.coverage_target < 0.0 or self.coverage_target > 100.0:
+                raise ConfigurationException(
+                    f'Invalid coverage target. {self.coverage_target} is not between 0 and 100.'
+                )
+        else:
+            self.coverage_target = None
 
     def _configure_group(self, group_name, group):
         """
@@ -143,22 +183,25 @@ class AnnotationConfig(object):
         """
         self.groups[group_name] = []
 
-        if not group:
-            raise TypeError('Group "{}" has no annotations!.'.format(group_name))
+        if not group or len(group) == 1:
+            raise ConfigurationException(f'Group "{group_name}" must have more than one annotation.')
 
         for annotation in group:
             for annotation_token in annotation:
                 annotation_value = annotation[annotation_token]
 
+                # Otherwise it should be a text type, if not then error out
+                if not self._is_annotation_token(annotation_value):
+                    raise ConfigurationException(f'{annotation} is an unknown annotation type.')
                 # The annotation comment is a choice group
                 if self._is_choice_group(annotation_value):
                     self._configure_choices(annotation_token, annotation_value)
-                # Otherwise it should be a text type, if not then error out
-                elif not self._is_annotation_token(annotation_value):
-                    raise TypeError('{} is an unknown annotation type.'.format(annotation))
+                # The annotation comment is not mandatory
+                if self._is_optional_group(annotation_value):
+                    self.optional_groups.append(annotation_token)
 
                 self.groups[group_name].append(annotation_token)
-                self.annotation_tokens.append(annotation_token)
+                self._add_annotation_token(annotation_token)
                 self.annotation_regexes.append(re.escape(annotation_token))
 
     def _configure_choices(self, annotation_token, annotation):
@@ -190,21 +233,20 @@ class AnnotationConfig(object):
 
             elif self._is_choice_group(annotation):
                 self._configure_choices(annotation_token_or_group_name, annotation)
-                self.annotation_tokens.append(annotation_token_or_group_name)
+                self._add_annotation_token(annotation_token_or_group_name)
                 self.annotation_regexes.append(re.escape(annotation_token_or_group_name))
 
             elif not self._is_annotation_token(annotation):  # pragma: no cover
-                # If not we don't know what to do with it
                 raise TypeError(
-                    '{} is an unknown type, must be strings or lists.'.format(annotation_token_or_group_name)
+                    f'{annotation_token_or_group_name} is an unknown type, must be strings or lists.'
                 )
             else:
-                self.annotation_tokens.append(annotation_token_or_group_name)
+                self._add_annotation_token(annotation_token_or_group_name)
                 self.annotation_regexes.append(re.escape(annotation_token_or_group_name))
 
-        self.echo.echo_v("Groups configured: {}".format(self.groups))
-        self.echo.echo_v("Choices configured: {}".format(self.choices))
-        self.echo.echo_v("Annotation tokens configured: {}".format(self.annotation_tokens))
+        self.echo.echo_v(f"Groups configured: {self.groups}")
+        self.echo.echo_v(f"Choices configured: {self.choices}")
+        self.echo.echo_v(f"Annotation tokens configured: {self.annotation_tokens}")
 
     def _plugin_load_failed_handler(self, *args, **kwargs):
         """
@@ -258,8 +300,7 @@ class AnnotationConfig(object):
             ))
 
 
-@six.add_metaclass(ABCMeta)
-class BaseSearch(object):
+class BaseSearch(metaclass=ABCMeta):
     """
     Base class for searchers.
     """
@@ -324,7 +365,7 @@ class BaseSearch(object):
         """
         # Not a choice type of annotation, nothing to do
         if annotation['annotation_token'] not in self.config.choices:
-            return
+            return None
 
         token = annotation['annotation_token']
         found_valid_choices = []
@@ -343,14 +384,29 @@ class BaseSearch(object):
                         )
                     )
                 elif choice in found_valid_choices:
-                    self._add_annotation_error(annotation, '"{}" is already present in this annotation.'.format(choice))
+                    self._add_annotation_error(annotation, f'"{choice}" is already present in this annotation.')
                 else:
                     found_valid_choices.append(choice)
         else:
             self._add_annotation_error(
                 annotation,
-                'No choices found for "{}". Expected one of {}.'.format(token, self.config.choices[token])
+                'no value found for "{}". Expected one of {}.'.format(token, self.config.choices[token])
             )
+        return None
+
+    def _get_group_children(self):
+        """
+        Create a list of all annotation tokens that are part of a group.
+
+        Returns:
+            List of annotation tokens that are configured to be in groups
+        """
+        group_children = []
+
+        for group in self.config.groups:
+            group_children.extend(self.config.groups[group])
+
+        return group_children
 
     def _get_group_for_token(self, token):
         """
@@ -365,6 +421,7 @@ class BaseSearch(object):
         for group in self.config.groups:
             if token in self.config.groups[group]:
                 return group
+        return None
 
     def check_results(self, all_results):
         """
@@ -378,78 +435,144 @@ class BaseSearch(object):
         Returns:
             Boolean indicating whether or not any errors were found
         """
-        if self.config.verbosity >= 2:
-            pprint.pprint(all_results, indent=3)
-
-        # This is used to quickly find out if a token is a member of a group
-        group_children = []
-
-        # Build a big list of all tokens that are part of a group
-        for group in self.config.groups:
-            group_children.extend(self.config.groups[group])
+        self.echo.pprint(all_results, indent=3, verbosity_level=2)
 
         # Spin through the search results
         for filename in all_results:
-            current_group = None
-            found_group_members = []
+            for annotations in self.iter_groups(all_results[filename]):
+                self.check_group(annotations)
+        return not self.errors
 
-            for annotation in all_results[filename]:
-                self._check_results_choices(annotation)
-                token = annotation['annotation_token']
+    def iter_groups(self, annotations):
+        """
+        Iterate on groups of annotations.
 
-                # TODO: Clean this up into reasonable methods
+        Annotations are considered as a group when they all have the same `line_number` and optional
+        `extra['object_id']`. The line number points to the beginning of the annotation group. The `object_id` is set
+        mostly for annotations parsed from a safelist.
+
+        Yield:
+            annotations (annotation list)
+        """
+        current_group = []
+        current_line_number = None
+        current_object_id = None
+        for annotation in annotations:
+            line_number = annotation["line_number"]
+            object_id = annotation.get("extra", {}).get("object_id")
+            line_number_changed = line_number != current_line_number
+            object_id_changed = object_id != current_object_id
+            if line_number_changed or object_id_changed:
                 if current_group:
-                    if token not in self.config.groups[current_group]:
-                        self._add_annotation_error(
-                            annotation,
-                            '"{}" is not in the group that starts with "{}". Expecting one of: {}'.format(
-                               token,
-                               current_group,
-                               self.config.groups[current_group]
-                            )
+                    yield current_group
+                current_group.clear()
+            current_group.append(annotation)
+            current_line_number = line_number
+            current_object_id = object_id
+
+        if current_group:
+            yield current_group
+
+    def check_group(self, annotations):
+        """
+        Perform several linting checks on a group of annotations.
+
+        The following checks are performed:
+
+        - Choice fields should have a valid value
+        - Annotation tokens are valid
+        - There is no duplicate
+        - All non-optional tokens are present
+        """
+        found_tokens = set()
+        group_tokens = []
+        group_name = None
+        for annotation in annotations:
+            token = annotation["annotation_token"]
+            if not group_name:
+                group_name = self._get_group_for_token(token)
+                if group_name:
+                    group_tokens = self.config.groups[group_name]
+
+            # Check if choice field
+            self._check_results_choices(annotation)
+
+            # Check token belongs to group
+            if group_name:
+                if token not in group_tokens:
+                    self._add_annotation_error(
+                        annotation,
+                        "'{}' token does not belong to group '{}'. Expected one of: {}".format(
+                           token,
+                           group_name,
+                           group_tokens
                         )
-                        current_group = None
-                        found_group_members = []
-                    elif token in found_group_members:
-                        self._add_annotation_error(
-                            annotation,
-                            '"{}" is already in the group that starts with "{}"'.format(token, current_group)
-                        )
-                        current_group = None
-                        found_group_members = []
-                    else:
-                        self.echo.echo_vv('Adding "{}", line {} to group {}'.format(
-                            token,
-                            annotation['line_number'],
-                            current_group
-                        ))
-                        found_group_members.append(token)
+                    )
 
-                        # If we have all members, this group is done
-                        if len(found_group_members) == len(self.config.groups[current_group]):
-                            self.echo.echo_vv("Group complete!")
-                            current_group = None
-                            found_group_members = []
-                else:
-                    if token in group_children:
-                        current_group = self._get_group_for_token(token)
+            # Check for duplicates
+            if token in found_tokens:
+                self._add_annotation_error(
+                    annotation,
+                    "found duplicate token '{}'".format(token)
+                )
+            if group_name:
+                found_tokens.add(token)
 
-                        if current_group is None:  # pragma: no cover
-                            # If we get here there is a problem with check_results' group_children not matching up with
-                            # our config's groups. That puts us in an unknown state, so we should quit.
-                            raise Exception(
-                                'group_children is out of sync with config.groups. {} is not in a group!'.format(token)
-                            )
+        # Check for missing tokens
+        for token in group_tokens:
+            if token not in self.config.optional_groups:
+                if token not in found_tokens:
+                    self._add_annotation_error(
+                        annotations[0],
+                        "missing non-optional annotation: '{}'".format(token)
+                    )
 
-                        found_group_members = [token]
-                        self.echo.echo_vv('Starting new group for "{}" token "{}", line {}'.format(
-                            current_group, token, annotation['line_number'])
-                        )
+    def check_annotation(self, annotation, current_group):
+        """
+        Check an annotation and add annotation errors when necessary.
 
+        Args:
+            annotation (dict): in particular, every annotation contains 'annotation_token' and 'annotation_data' keys.
+            current_group (str): None or the name of a group (from self.config.groups) to which preceding annotations
+                belong.
+            found_group_tokens (list): annotation tokens from the same group that were already found. This list is
+                cleared in case of error or when creating a new group.
+
+        Return:
+            current_group (str or None)
+        """
+        self._check_results_choices(annotation)
+        token = annotation['annotation_token']
+
+        if current_group:
+            # Add to existing group
+            if token not in self.config.groups[current_group]:
+                # Check for token correctness
+                self._add_annotation_error(
+                    annotation,
+                    '"{}" is not in the group that starts with "{}". Expecting one of: {}'.format(
+                       token,
+                       current_group,
+                       self.config.groups[current_group]
+                    )
+                )
+                current_group = None
+            else:
+                # Token is correct
+                self.echo.echo_vv('Adding "{}", line {} to group {}'.format(
+                    token,
+                    annotation['line_number'],
+                    current_group
+                ))
+        else:
+            current_group = self._get_group_for_token(token)
             if current_group:
-                self.errors.append('File finished with an incomplete group {}!'.format(current_group))
+                # Start a new group
+                self.echo.echo_vv('Starting new group for "{}" token "{}", line {}'.format(
+                    current_group, token, annotation['line_number'])
+                )
 
-        return False if self.errors else True
+        return current_group
 
     def _add_annotation_error(self, annotation, message):
         """
@@ -482,14 +605,81 @@ class BaseSearch(object):
         Returns:
             Dict of {filename: annotations} for all files with found annotations.
         """
-        pass  # pragma: no cover
 
-    def report(self, all_results):
+    def _format_results_for_report(self, all_results):
+        """
+        Format the given results dict for reporting purposes.
+
+        Args:
+            all_results: Dict of all results found in a search
+
+        Returns:
+            Dict of results arranged for reporting
+        """
+        group_children = self._get_group_children()
+        formatted_results = {}
+        current_group_id = 0
+
+        for filename in all_results:
+            self.echo.echo_vv(f"report_format: formatting {filename}")
+            formatted_results[filename] = []
+            current_group = None
+
+            found_group_members = []
+
+            for annotation in all_results[filename]:
+                token = annotation['annotation_token']
+                self.echo.echo_vvv(f"report_format: formatting annotation token {token}")
+
+                if current_group:
+                    if token not in self.config.groups[current_group]:
+                        self.echo.echo_vv(
+                            "report_format: {} is not a group member, finishing group id {}".format(
+                                token,
+                                current_group_id
+                            )
+                        )
+                        current_group = None
+                        found_group_members = []
+                        formatted_results[filename].append(annotation)
+                    else:
+                        self.echo.echo_vv("report_format: Adding {} to group id {}".format(
+                            token,
+                            current_group_id
+                        ))
+                        annotation['report_group_id'] = current_group_id
+                        formatted_results[filename].append(annotation)
+                        found_group_members.append(token)
+                else:
+                    if token in group_children:
+                        current_group = self._get_group_for_token(token)
+                        current_group_id += 1
+                        found_group_members = [token]
+                        annotation['report_group_id'] = current_group_id
+                        formatted_results[filename].append(annotation)
+
+                        self.echo.echo_vv('Starting group id {} for "{}" token "{}", line {}'.format(
+                            current_group_id, current_group, token, annotation['line_number'])
+                        )
+                    else:
+                        self.echo.echo_vv(f'Adding single token {token}.')
+                        formatted_results[filename].append(annotation)
+
+                # If we have all members, this group is done
+                if current_group and len(found_group_members) == len(self.config.groups[current_group]):
+                    self.echo.echo_vv("report_format: Group complete!")
+                    current_group = None
+                    found_group_members = []
+
+        return formatted_results
+
+    def report(self, all_results, report_prefix=''):
         """
         Genrates the YAML report of all search results.
 
         Args:
             all_results: Dict of found annotations, indexed by filename
+            report_prefix: Prefix to add to report filename
 
         Returns:
             Filename of generated report
@@ -497,9 +687,13 @@ class BaseSearch(object):
         self.echo.echo_vv(yaml.dump(all_results, default_flow_style=False))
 
         now = datetime.datetime.now()
-        report_filename = os.path.join(self.config.report_path, '{}.yaml'.format(now.strftime('%Y-%d-%m-%H-%M-%S')))
+        report_filename = os.path.join(self.config.report_path, '{}{}.yaml'.format(
+            report_prefix, now.strftime('%Y-%d-%m-%H-%M-%S')
+        ))
 
-        self.echo("Generating report to {}".format(report_filename))
+        formatted_results = self._format_results_for_report(all_results)
+
+        self.echo(f"Generating report to {report_filename}")
 
         try:
             os.makedirs(self.config.report_path)
@@ -508,6 +702,6 @@ class BaseSearch(object):
                 raise
 
         with open(report_filename, 'w+') as report_file:
-            yaml.dump(all_results, report_file, default_flow_style=False)
+            yaml.safe_dump(formatted_results, report_file, default_flow_style=False)
 
         return report_filename
